import json
import logging
import os
import sys
import time
from abc import ABC
from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import socketio
import ujson
from lz4.frame import compress, decompress

from era_5g_interface.exceptions import BackPressureException, UnknownChannelTypeUsed
from era_5g_interface.frame_decoder import FrameDecoder, FrameDecoderError
from era_5g_interface.frame_encoder import FrameEncoder, FrameEncoderError
from era_5g_interface.measuring import Measuring

logger = logging.getLogger(__name__)

# TODO: use enums?
DATA_NAMESPACE = str("/data")
DATA_ERROR_EVENT = str("data_error")

CONTROL_NAMESPACE = str("/control")
COMMAND_EVENT = str("command")
COMMAND_ERROR_EVENT = str("command_error")
COMMAND_RESULT_EVENT = str("command_result")


@unique
class ChannelType(Enum):
    """Channel type dataclass."""

    JSON = 1
    JPEG = 2
    H264 = 3
    JSON_LZ4 = 4
    HEVC = 5


@dataclass
class CallbackInfoClient:
    """Callback info dataclass used on client side."""

    type: ChannelType
    callback: Callable[[Dict], Any]  # Custom callback with dict.
    error_event: str = DATA_ERROR_EVENT  # Custom error event name.


@dataclass
class CallbackInfoServer:
    """Callback info dataclass used on server side - callback has namespace sid parameter."""

    type: ChannelType
    callback: Callable[[str, Dict], Any]  # Custom callback with sid and dict.
    error_event: str = DATA_ERROR_EVENT  # Custom error event name.


class Channels(ABC):
    """Channels class is used to define channel data callbacks and contains send functions.

    It handles image frames JPEG, H.264 and HEVC encoding/decoding. Data is sent via the DATA_NAMESPACE. The class
    cannot be used alone. The ServerChannels and ClientChannels classes create callbacks and encoders/decoders.
    """

    # This should work roughly like an abstract member.
    _callbacks_info: Union[Dict[str, CallbackInfoClient], Dict[str, CallbackInfoServer]]

    def __init__(
        self,
        sio: Union[socketio.Client, socketio.Server],
        callbacks_info: Union[Dict[str, CallbackInfoClient], Dict[str, CallbackInfoServer]],
        disconnect_callback: Optional[Callable] = None,
        back_pressure_size: Optional[int] = 5,
        recreate_coder_attempts_count: int = 5,
        stats: bool = False,
        extended_measuring: bool = False,
    ):
        """Constructor.

        Args:
            sio (Union[socketio.Client, socketio.Server]): Socketio Client or Server object.
            callbacks_info (Union[Dict[str, CallbackInfoClient], Dict[str, CallbackInfoServer]]): Callbacks Info
                dictionary, key is custom event name.
            disconnect_callback (Callable, optional): Triggered before _shutdown on unhandled exception.
            back_pressure_size (int, optional): Back pressure size - max size of eio.queue.qsize().
            recreate_coder_attempts_count (int): How many times try to recreate the frame encoder/decoder.
            stats (bool): Store output data sizes.
            extended_measuring (bool): Enable logging of measuring.
        """

        self._sio = sio
        self._disconnect_callback = disconnect_callback

        if back_pressure_size is not None and back_pressure_size < 1:
            raise ValueError("Invalid value for back_pressure_size.")

        self._back_pressure_size = back_pressure_size
        self._recreate_coder_attempts_count = recreate_coder_attempts_count
        self._stats = stats
        if self._stats:
            self._sizes: List[int] = []
        self._extended_measuring = extended_measuring

        if isinstance(self._sio, socketio.Client):
            measuring_filename_prefix = "client-"
        else:
            measuring_filename_prefix = "server-"

        self._input_measuring = Measuring(
            measuring_items={
                "key_timestamp": 0,
                "eio_sid": 0,
                "event": "",
                "before_decode_timestamp": 0,
                "after_decode_timestamp": 0,
                "before_lz4_decode_timestamp": 0,
                "after_lz4_decode_timestamp": 0,
                "before_callback_timestamp": 0,
                "after_callback_timestamp": 0,
            },
            enabled=self._extended_measuring,
            filename_prefix=measuring_filename_prefix + "channels-input",
        )

        self._output_measuring = Measuring(
            measuring_items={
                "key_timestamp": 0,
                "eio_sid": 0,
                "event": "",
                "worker_recv_timestamp": 0,
                "worker_before_process_timestamp": 0,
                "worker_after_process_timestamp": 0,
                "worker_send_timestamp": 0,
                "before_encode_timestamp": 0,
                "after_encode_timestamp": 0,
                "size": 0,
                "before_lz4_encode_timestamp": 0,
                "after_lz4_encode_timestamp": 0,
                "before_send_timestamp": 0,
                "after_send_timestamp": 0,
            },
            enabled=self._extended_measuring,
            filename_prefix=measuring_filename_prefix + "channels-output",
        )

        self._callbacks_info = callbacks_info
        # For multiple frame streams, the encoders and the decoders are indexed by Tuple(eio_sid, event).
        self._decoders: Dict[Tuple[str, str], FrameDecoder] = dict()
        self._encoders: Dict[Tuple[str, str], FrameEncoder] = dict()

    def _shutdown(self, cb_type: str, event: str, sid: Optional[str] = None) -> None:
        logger.error(f"Unhandled exception in {cb_type} callback (event: {event}).", exc_info=sys.exc_info())
        if self._disconnect_callback:
            if isinstance(self._sio, socketio.Client):
                self._disconnect_callback()
            else:
                self._disconnect_callback(sid, DATA_NAMESPACE)
        logging.shutdown()  # should flush the logger
        os._exit(1)  # standard sys.exit() is not enough

    def _apply_back_pressure(self, sid: Optional[str] = None) -> None:
        """Apply back pressure."""

        if self._back_pressure_size is not None:
            if isinstance(self._sio, socketio.Client):
                if self._sio.eio.queue.qsize() > self._back_pressure_size:
                    raise BackPressureException()
            else:
                if sid is None:
                    raise ValueError("'sid' has to be set for server.")
                eio_sid = self.get_client_eio_sid(str(sid), DATA_NAMESPACE)
                if self._sio.eio.sockets[eio_sid].queue.qsize() > self._back_pressure_size:
                    raise BackPressureException()

    @staticmethod
    def get_timestamp_from_data(data: Dict, name: str = "timestamp", default: Optional[int] = None):
        if default is None:
            default = time.perf_counter_ns()
        timestamp = data.get(name, 0)
        if not timestamp:
            try:
                timestamp = json.loads(data.get("data", "")).get(name, default)
            except ValueError:
                timestamp = default
        return timestamp

    def send_image(
        self,
        frame: np.ndarray,
        event: str,
        channel_type: ChannelType,
        timestamp: Optional[int] = None,
        metadata: Optional[Dict[str, Any]] = None,
        sid: Optional[str] = None,
        can_be_dropped: bool = True,
        wait_for_reconnection: bool = True,
        blocking: bool = False,
        encoding_options: Optional[Dict[str, str]] = None,
    ) -> Any:
        """Send general image data with JPEG or H.264 or HEVC encoding via DATA_NAMESPACE.

        NOTE: DATA_NAMESPACE is assumed to be a connected namespace.

        Args:
            frame (np.ndarray): Video frame / image.
            event (str): Event name.
            channel_type (ChannelType): Encoding type - ChannelType.JPEG or ChannelType.H264 or ChannelType.HEVC.
            timestamp (int): Frame timestamp.
            metadata (Dict[str, Any], optional): Optional metadata to send.
            sid (str, optional): Namespace sid - mandatory when sending from the server side to the client.
            can_be_dropped (bool): If data can be lost due to back pressure.
            wait_for_reconnection (bool): Wait for reconnection to server side? Wait is blocking.
            blocking (bool): If True, wait for the response. Defaults to False.
            encoding_options (Dict[str, str], optional): Video codec options, e.g. {"crf": "0", "preset":
                "ultrafast", "tune": "zerolatency", "x264-params": "keyint=5"}, default: {"preset": "ultrafast",
                "tune": "zerolatency"}.

        Parameter frame.shape should not be changed after first send_image function call.
        Parameters encoding_options and frame.shape are only used in the first send_image call to create the encoder.
        """

        if channel_type not in (ChannelType.JPEG, ChannelType.H264, ChannelType.HEVC):
            raise UnknownChannelTypeUsed()

        if timestamp is None:
            timestamp = time.perf_counter_ns()
        data: Dict[str, Any] = {"timestamp": timestamp}
        if metadata:
            data["metadata"] = metadata
        eio_sid = self.get_client_eio_sid(sid, DATA_NAMESPACE)
        encoder_id = (eio_sid, event)
        if channel_type in (ChannelType.H264, ChannelType.HEVC):
            if encoder_id not in self._encoders:
                try:
                    if channel_type is ChannelType.H264:
                        logger.info(f"Creating H.264 encoder for image size {frame.shape[1]}x{frame.shape[0]}")
                        self._encoders[encoder_id] = FrameEncoder(
                            frame.shape[1], frame.shape[0], codec="h264", options=encoding_options
                        )
                    elif channel_type is ChannelType.HEVC:
                        logger.info(f"Creating HEVC encoder for image size {frame.shape[1]}x{frame.shape[0]}")
                        self._encoders[encoder_id] = FrameEncoder(
                            frame.shape[1], frame.shape[0], codec="hevc", options=encoding_options
                        )
                except Exception as e:
                    logger.error(f"Cannot create video frame encoder: {repr(e)}")
                    raise e
        try:
            is_key_frame = False

            self._output_measuring.log_timestamp(timestamp, "before_encode_timestamp")

            if channel_type in (ChannelType.H264, ChannelType.HEVC):
                frame_encoded = self._encoders[encoder_id].encode_ndarray(frame)
                # TODO: dataclass for this data
                data["width"] = self._encoders[encoder_id].width()
                data["height"] = self._encoders[encoder_id].height()
                is_key_frame = self._encoders[encoder_id].last_frame_is_keyframe()
            else:
                _, frame_jpeg = cv2.imencode(".jpg", frame)
                frame_encoded = frame_jpeg.tobytes()

            self._output_measuring.log_timestamp(timestamp, "after_encode_timestamp")

            data["frame"] = frame_encoded

            if self._stats:
                # TODO: include all data size
                self._sizes.append(len(frame_encoded))
                logger.debug(f"Frame data size: {self._sizes[-1]}")

                self._output_measuring.log_measuring(timestamp, "size", self._sizes[-1])

            if len(frame_encoded):
                self.send_data(
                    data,
                    event,
                    sid=sid,
                    can_be_dropped=(can_be_dropped and not is_key_frame),
                    wait_for_reconnection=wait_for_reconnection,
                    blocking=blocking,
                )
            else:
                logger.warning(
                    "Encoded frame size is zero. It is possible that the encoder settings allow a delay "
                    "of several frames at the beginning of the stream. This means that the timestamps "
                    "will be shifted and will not match the original frames."
                )
        except FrameEncoderError as e:
            logger.error(f"Video frame encoder error: {e}")
            # Try to recreate encoder
            if self._encoders[encoder_id].get_init_count() < self._recreate_coder_attempts_count:
                logger.info(f"Try to recreate encoder ... attempt {self._encoders[encoder_id].get_init_count()}")
                self._encoders[encoder_id].encoder_init()
            else:
                raise e

    def send_data(
        self,
        data: Dict[str, Any],
        event: str,
        channel_type: ChannelType = ChannelType.JSON,
        sid: Optional[str] = None,
        can_be_dropped: bool = False,
        wait_for_reconnection: bool = True,
        blocking: bool = False,
    ) -> Any:
        """Send general JSON data via DATA_NAMESPACE.

        NOTE: DATA_NAMESPACE is assumed to be a connected namespace.

        Args:
            data (Dict[str, Any]): JSON data.
            event (str): Event name.
            channel_type (ChannelType): ChannelType.JSON for raw JSON or ChannelType.JSON_LZ4 for LZ4 compressed JSON.
            sid (str, optional): Namespace sid - mandatory when sending from the server side to the client.
            can_be_dropped (bool): If data can be lost due to back pressure.
            wait_for_reconnection (bool): Wait for reconnection to server side? Wait is blocking.
            blocking (bool): If True, wait for the response. Defaults to False.
        """

        if channel_type is not ChannelType.JSON and channel_type is not ChannelType.JSON_LZ4:
            raise UnknownChannelTypeUsed()

        timestamp = Channels.get_timestamp_from_data(data)

        self._output_measuring.log_measuring(
            timestamp, "worker_recv_timestamp", Channels.get_timestamp_from_data(data, "recv_timestamp", 0)
        )
        self._output_measuring.log_measuring(
            timestamp,
            "worker_before_process_timestamp",
            Channels.get_timestamp_from_data(data, "timestamp_before_process", 0),
        )
        self._output_measuring.log_measuring(
            timestamp,
            "worker_after_process_timestamp",
            Channels.get_timestamp_from_data(data, "timestamp_after_process", 0),
        )
        self._output_measuring.log_measuring(
            timestamp, "worker_send_timestamp", Channels.get_timestamp_from_data(data, "send_timestamp", 0)
        )

        if can_be_dropped:
            self._apply_back_pressure(sid)

        new_data = data
        if channel_type is ChannelType.JSON_LZ4:
            self._output_measuring.log_timestamp(timestamp, "before_lz4_encode_timestamp")
            new_data = compress(bytes(ujson.dumps(data), "utf-8"))
            self._output_measuring.log_timestamp(timestamp, "after_lz4_encode_timestamp")

        self._output_measuring.log_timestamp(timestamp, "before_send_timestamp")

        if isinstance(self._sio, socketio.Client):
            if wait_for_reconnection:
                while not self._sio.connected or not self._sio.eio.state:
                    logger.info("Waiting for reconnection ...")
                    if not self._sio._reconnect_task or not self._sio._reconnect_task.is_alive():
                        break
                    time.sleep(1)
            if blocking:
                return self._sio.call(event, new_data, namespace=DATA_NAMESPACE)
            else:
                self._sio.emit(event, new_data, namespace=DATA_NAMESPACE)
        else:
            if sid is None:
                raise ValueError("'sid' has to be set for server.")
            if not self._sio.manager.is_connected(sid, DATA_NAMESPACE):
                raise ConnectionError(f"Client with {DATA_NAMESPACE} sid {sid} is not connected to server.")
            if blocking:
                self._sio.call(event, new_data, namespace=DATA_NAMESPACE, to=sid)
            else:
                self._sio.emit(event, new_data, namespace=DATA_NAMESPACE, to=sid)

        self._output_measuring.log_timestamp(timestamp, "after_send_timestamp")
        self._output_measuring.log_measuring(timestamp, "eio_sid", self.get_client_eio_sid(sid, DATA_NAMESPACE))
        self._output_measuring.log_measuring(timestamp, "event", event)
        self._output_measuring.store_measuring(timestamp)

    def get_client_eio_sid(self, sid: Optional[str] = None, namespace: Optional[str] = None) -> str:
        """Get client eio sid.

        Args:
            sid (str, optional): Namespace sid - mandatory for using on the server side.
            namespace (str, optional): Namespace - mandatory for using on the server side.

        Returns:
            Client eio sid.
        """

        if isinstance(self._sio, socketio.Client):
            return str(self._sio.sid)
        else:
            if sid is None:
                raise ValueError("'sid' has to be set for server.")
            return str(self._sio.manager.eio_sid_from_sid(sid, namespace))

    def data_error_callback(self, data: Dict[str, Any], sid: Optional[str] = None) -> None:
        """Allows to receive general error data on DATA_NAMESPACE.

        Args:
            data (Dict[str, Any]): JSON data.
            sid (str, optional): Namespace sid - only on the server side.
        """

        logger.error(f"Data error, eio_sid {self.get_client_eio_sid(sid, DATA_NAMESPACE)}, sid {sid}, data {data}")

    def image_decode(self, data: Dict[str, Any], event: str, sid: Optional[str] = None) -> Optional[Dict]:
        """Decode JPEG or H.264 or HEVC encoded image received on DATA_NAMESPACE.

        Args:
            data (Dict[str, Any]): Received dictionary with frame data.
            event (str): Event name.
            sid (str, optional): Namespace sid - only on the server side.
        """

        if "timestamp" not in data:
            logger.error("Missing timestamp in frame data dictionary.")
            self.send_data(
                {"timestamp": 0, "error": "Missing timestamp in frame data dictionary."},
                self._callbacks_info[event].error_event,
                sid=sid,
            )
            return None
        timestamp = data.get("timestamp", 0)

        if "frame" not in data:
            logger.error("Data does not contain frame.")
            self.send_data(
                {"timestamp": timestamp, "error": "Data does not contain frame."},
                self._callbacks_info[event].error_event,
                sid=sid,
            )
            return None

        eio_sid = self.get_client_eio_sid(sid, DATA_NAMESPACE)
        decoder_id = (eio_sid, event)
        if (
            self._callbacks_info[event].type in (ChannelType.H264, ChannelType.HEVC)
            and decoder_id not in self._decoders
        ):
            if "width" not in data or "height" not in data:
                logger.error("Data does not contain width or height, it is mandatory for video frame decoder.")
                self.send_data(
                    {
                        "timestamp": timestamp,
                        "error": "Data does not contain width or height, it is mandatory for video frame decoder.",
                    },
                    self._callbacks_info[event].error_event,
                    sid=sid,
                )
                return None
            try:
                if self._callbacks_info[event].type is ChannelType.H264:
                    logger.info(f"Creating H.264 decoder for image size {data['width']}x{data['height']}")
                    self._decoders[decoder_id] = FrameDecoder(data["width"], data["height"], codec="h264")
                elif self._callbacks_info[event].type is ChannelType.HEVC:
                    logger.info(f"Creating HEVC decoder for image size {data['width']}x{data['height']}")
                    self._decoders[decoder_id] = FrameDecoder(data["width"], data["height"], codec="hevc")
            except Exception as e:
                logger.error(f"Cannot create video frame decoder: {repr(e)}")
                self.send_data(
                    {"timestamp": timestamp, "error": f"Cannot create video frame decoder: {repr(e)}"},
                    self._callbacks_info[event].error_event,
                    sid=sid,
                )
                return None

        if decoder_id in self._decoders:
            last_timestamp = self._decoders[decoder_id].last_timestamp
            if timestamp - last_timestamp < 0:
                logger.error(
                    f"Received frame with older timestamp: {timestamp}, "
                    f"last_timestamp: {last_timestamp}, diff: {timestamp - last_timestamp}"
                )
                self.send_data(
                    {
                        "timestamp": timestamp,
                        "error": (
                            f"Received frame with older timestamp: {timestamp}, "
                            f"last_timestamp: {last_timestamp}, diff: {timestamp - last_timestamp}"
                        ),
                    },
                    self._callbacks_info[event].error_event,
                    sid=sid,
                )
                return None
            self._decoders[decoder_id].last_timestamp = timestamp

        if decoder_id in self._decoders:
            try:
                frame_decoded = self._decoders[decoder_id].decode_packet_data(data["frame"])
            except FrameDecoderError as e:
                logger.error(f"Video frame decoder error: {e}")
                # Try to recreate decoder
                if self._decoders[decoder_id].get_init_count() < self._recreate_coder_attempts_count:
                    logger.info(f"Try to recreate decoder ... attempt {self._decoders[decoder_id].get_init_count()}")
                    # self._decoders[decoder_id].decoder_init()
                self.send_data(
                    {"timestamp": timestamp, "error": f"Video frame decoder error: {e}"},
                    self._callbacks_info[event].error_event,
                    sid=sid,
                )
                return None
        else:
            try:
                frame_decoded = cv2.imdecode(np.frombuffer(data["frame"], dtype=np.uint8), cv2.IMREAD_COLOR)
            except Exception as e:
                logger.error(f"Failed to decode frame data: {repr(e)}")
                self.send_data(
                    {"timestamp": timestamp, "error": f"Failed to decode frame data: {repr(e)}"},
                    self._callbacks_info[event].error_event,
                    sid=sid,
                )
                return None

        decoded_data = {"frame": frame_decoded, "timestamp": timestamp}
        if "metadata" in data:
            decoded_data["metadata"] = data["metadata"]

        return decoded_data

    def data_lz4_decode(self, data: bytes, event: str, sid: Optional[str] = None) -> Optional[Dict]:
        """Decode LZ4 compressed general JSON data received on DATA_NAMESPACE.

        Args:
            data (bytes): LZ4 compressed JSON data.
            event (str): Event name.
            sid (str, optional): Namespace sid - only on the server side.
        """

        assert isinstance(data, bytes)

        try:
            new_data: Dict = ujson.loads(decompress(data))
            return new_data
        except Exception as e:
            logger.error(f"Failed to decode LZ4 JSON data: {repr(e)}")
            self.send_data(
                {"error": f"Failed to decode LZ4 JSON data: {repr(e)}"},
                self._callbacks_info[event].error_event,
                sid=sid,
            )
            return None

    @property
    def stats(self):
        return self._stats

    @property
    def sizes(self):
        return self._sizes



class Era5gInterfaceException(Exception):
    """Common base for all exception that this client might raise."""

    pass


class FailedToSendData(Era5gInterfaceException):
    """Raised when the data could not be sent."""

    pass


class BackPressureException(Era5gInterfaceException):
    """Raised when sending too much data (output buffer too filled)."""

    pass


class UnknownChannelTypeUsed(Era5gInterfaceException):
    """Raised when unknown ChannelType is used."""

    pass



import csv
import logging
import time
from datetime import datetime
from typing import Any, Dict

logger = logging.getLogger(__name__)


class Measuring:
    """Measuring class is used to define and measure and log timestamps or custom data to csv file."""

    def __init__(
        self,
        measuring_items: Dict[str, Any],
        enabled: bool = False,
        filename_prefix: str = "measuring",
    ):
        """Constructor."""

        self._enabled = enabled
        if self._enabled:
            self._measuring: Dict[int, Any] = {}
            self._measuring_file = open(
                "{}-{}.csv".format(filename_prefix, datetime.now().strftime("%Y%m%d_%H%M%S_%f")[:-3]),
                "w",
                newline="\n",
            )
            self._csv_writer = csv.writer(self._measuring_file)
            self._measuring_items = measuring_items
            if "key_timestamp" not in self._measuring_items:
                self._measuring_items["key_timestamp"] = 0
            self._csv_writer.writerow(self._measuring_items.keys())
            self._measuring_file.flush()

    def log_measuring(self, key_timestamp: int, key: str, value: Any) -> None:
        if self._enabled:
            if key_timestamp not in self._measuring:
                self._measuring[key_timestamp] = self._measuring_items.copy()
                self._measuring[key_timestamp]["key_timestamp"] = key_timestamp
            self._measuring[key_timestamp][key] = value

    def log_timestamp(self, key_timestamp: int, key: str) -> None:
        self.log_measuring(key_timestamp, key, time.perf_counter_ns())

    def store_measuring(self, key_timestamp: int) -> None:
        if self._enabled:
            self._csv_writer.writerow(self._measuring[key_timestamp].values())
            self._measuring_file.flush()



import logging
from enum import IntEnum
from queue import Empty, Full, Queue
from threading import Lock
from typing import Any, Dict, Optional

from era_5g_interface.dataclasses.control_command import ControlCommand

logger: logging.Logger = logging.getLogger(__name__)


class QueueFullAction(IntEnum):
    """Action type for ful queue."""

    DISCARD_OLDEST = 1
    DISCARD_NEW = 2
    RAISE = 3


class QueueDiscardError(Exception):
    """Exception which is raised when a problem occurs during discarding of old data when internal queue is full."""

    pass


class TaskHandlerInternalQ:
    """Task handler which takes care of passing the data to the python internal queue for future processing.

    It could either be inherited to implement the _run method and read the data from any source or used directly and
    call the store_data method externally.
    """

    def __init__(
        self,
        data_queue: Queue,
        if_queue_full: QueueFullAction = QueueFullAction.DISCARD_OLDEST,
        queue_put_lock: Optional[Lock] = None,
    ) -> None:
        """Constructor.

        Args:
            data_queue (Queue): The queue where the data and metadata should be passed to.
            if_queue_full (QueueFullAction): This parameter defines what should happen in case the internal
                queue is full and new data is to be added. Possible actions are defined by QueueFullAction Enum:
                DISCARD_OLDEST, DISCARD_NEW and RAISE. Default is DISCARD_OLDEST, which removes the oldest item from
                the queue to make up space for the new one. Option DISCARD_NEW throws away the newly provided data
                and keeps the queue intact. Choosing RAISE results in raising RuntimeError exception. Note that even
                DISCARD_OLDEST can result in discarding the new data - this happens in case the oldest data is a
                Control Command. Note also that, since Control Commands are considered more important than regular
                data, attempting to discard new Control Command raises an exception, even in case the value is set to
                DISCARD_NEW.
            queue_put_lock (Lock, optional): Lock for preventing multiple task handlers (or calls to store_data from
                multiple threads) from adding new data during the possible discarding of old data when queue is full.
                Default is None (the lock will be created internally).
        """

        self._q: Queue = data_queue
        assert isinstance(if_queue_full, QueueFullAction)
        self.if_queue_full = if_queue_full
        if queue_put_lock is not None:
            self._q_put_lock = queue_put_lock
        else:
            self._q_put_lock = Lock()

    def data_queue_size(self) -> int:
        """Data queue size.

        Returns:
            Data queue size
        """

        return self._q.qsize()

    def data_queue_occupancy(self) -> float:
        """Data queue occupancy.

        Returns:
            Data queue occupancy.
        """

        return self._q.qsize() / self._q.maxsize

    def store_data(self, metadata: Dict, data: Any) -> None:
        """Method which will store the data (e.g. an image) to the queue for processing.

        Args:
            metadata (Dict): Arbitrary dictionary with metadata related to the data. The format is 5G-ERA Network
                Application specific.
            data (Any): The data to be processed.

        Raises:
            QueueDiscardError: Raised when a problem occurs during discarding of old data when internal queue is full.
        """

        # Make sure multiple calls of store_data do not interfere with each other.
        with self._q_put_lock:
            try:
                self._q.put((metadata, data), block=False)
            except Full:
                if self.if_queue_full == QueueFullAction.DISCARD_NEW:
                    pass  # Do nothing (discard the new data).
                elif self.if_queue_full == QueueFullAction.DISCARD_OLDEST:
                    # Remove the oldest data from queue to make up space for the new one,
                    # but make sure not to discard control commands.
                    space_available = False
                    with self._q.mutex:  # Acquire the queue lock to prevent it from mutating during the removal.
                        if isinstance(self._q.queue[0], ControlCommand):
                            logger.warning(
                                "Discarding new data instead of the old one to prevent losing a control command in "
                                "queue."
                            )
                            # This situation could also raise an exception, if desired.
                        else:
                            self._q._get()  # Remove the old data from queue.
                            space_available = True

                    if space_available:
                        self._q.put((metadata, data), block=False)
                else:
                    raise QueueDiscardError("Internal queue is full and no data was allowed to be discarded.")

    def store_control_data(self, data: ControlCommand) -> None:
        """Pass control commands to the worker using internal queue.

        Args:
            data (ControlCommand): ControlCommand with control data.

        Raises:
            QueueDiscardError: Raised when a problem occurs during discarding of old data when internal queue is full.
        """

        with self._q_put_lock:
            try:
                self._q.put(data, block=False)
            except Full:
                if self.if_queue_full == QueueFullAction.DISCARD_OLDEST:
                    with self._q.mutex:
                        if isinstance(self._q.queue[0], ControlCommand):
                            raise QueueDiscardError(
                                "Internal queue is full, and the new Control Command would discard a previous one."
                            )
                        else:
                            self._q._get()  # remove the old data from queue

                    self._q.put(data, block=False)
                else:
                    raise QueueDiscardError("Internal queue is full, Control Command cannot be processed.")

    def clear_storage(self) -> None:
        """Clear all items from internal queue."""

        while not self._q.empty():
            try:
                self._q.get(block=False)
            except Empty:
                break
            self._q.task_done()



import logging

import numpy as np
from av.codec import CodecContext
from av.error import FFmpegError, tag_to_code
from av.packet import Packet
from av.video.codeccontext import VideoCodecContext
from av.video.frame import VideoFrame


class FrameDecoderError(Exception):
    """FrameDecoderError Exception."""

    pass


# TODO: only for testing purpose
# from pathlib import Path
# Path("output").mkdir(parents=True, exist_ok=True)

logger = logging.getLogger("Video frame decoder")


class FrameDecoder:
    """Video frame Decoder."""

    def __init__(self, width: int, height: int, fps: float = 30, codec: str = "h264") -> None:
        """Constructor.

        Args:
            width (int): Video frame width.
            height (int): Video frame height.
            fps (float): Video framerate (FPS), default: 30.
            codec (str): Video codec name, e.g. h264, hevc, or vp9, default: h264.
        """
        self._fps = fps
        self._width = width
        self._height = height
        self._pix_fmt = "yuv420p"
        self._decoder: VideoCodecContext = CodecContext.create(codec, "r")
        self._init_count = 0
        self.last_timestamp: int = 0
        self._last_frame_is_keyframe = False
        self.decoder_init()

    def width(self) -> int:
        """Get video frame width.

        Returns:
            Video frame width.
        """

        return self._width

    def height(self) -> int:
        """Get video frame height.

        Returns:
            Video frame height.
        """

        return self._height

    def fps(self) -> float:
        """Get video framerate.

        Returns:
            Video framerate.
        """

        return self._fps

    def decoder_init(self) -> None:
        """Init video decoder."""

        self._init_count += 1
        if self._decoder.is_open:
            self._decoder.close()
        self._decoder.width = self._width
        self._decoder.height = self._height
        self._decoder.framerate = self._fps
        self._decoder.pix_fmt = self._pix_fmt
        self._decoder.open()

    def get_init_count(self) -> int:
        """Get decoder init attempts count.

        Returns:
            Decoder init attempts count.
        """

        return self._init_count

    def last_frame_is_keyframe(self) -> bool:
        """Is last frame a keyframe?

        Returns:
            True if last frame is keyframe.
        """

        return self._last_frame_is_keyframe

    def decode_packet_data(self, packet_data: bytes, format: str = "bgr24") -> np.ndarray:
        """Decode packets bytes to ndarray.

        Args:
            packet_data (bytes): Packet data.
            format (str): Image format.

        Returns:
            Video frame / image.
        """

        try:
            packet = Packet(packet_data)

            # Multiple frames? - This should not happen because on the encoders side one frame is always encoded and sent
            frame: VideoFrame
            for frame in self._decoder.decode(packet):
                # TODO: only for testing purpose
                logger.debug(f"Frame {frame} with id {frame.index} decoded from packet: {packet}")
                logger.debug(
                    f"frame.pts: {frame.pts}, "
                    f"frame.dts: {frame.dts}, "
                    f"frame.index: {frame.index}, "
                    f"frame.key_frame: {frame.key_frame}, "
                    f"frame.is_corrupt: {frame.is_corrupt}, "
                    f"frame.time: {frame.time}"
                )
                # TODO: only for testing purpose
                # frame.to_image().save("output/frame-%04d.jpg" % frame.index)

                self._last_frame_is_keyframe = frame.key_frame
                frame_ndarray: np.ndarray = frame.to_ndarray(format=format)
                return frame_ndarray
            # Sometimes, with dropping some TCP packets, no frame is decoded from av.Packet while no FFmpegError is
            # raised. This is resolved by throwing this exception.
            raise FFmpegError(tag_to_code(b"INDA"), f"No frame decoded from packet {packet}")
        except FFmpegError as e:
            raise FrameDecoderError(e)



from threading import Lock
from typing import Any, Iterator


class LockedSet(set):
    """A set which can be safely iterated and modified from different threads."""

    def __init__(self, *args, **kwargs):
        self._lock = Lock()
        super(LockedSet, self).__init__(*args, **kwargs)

    def add(self, elem: Any):
        with self._lock:
            super(LockedSet, self).add(elem)

    def remove(self, elem: Any):
        with self._lock:
            super(LockedSet, self).remove(elem)

    def discard(self, elem: Any) -> None:
        with self._lock:
            return super(LockedSet, self).remove(elem)

    def __contains__(self, elem: Any):
        with self._lock:
            super(LockedSet, self).__contains__(elem)

    def locked_iter(self, it: Iterator) -> Iterator:
        it = iter(it)

        while True:
            try:
                with self._lock:
                    value = next(it)
            except StopIteration:
                return
            yield value

    def __iter__(self) -> Iterator:
        return self.locked_iter(super().__iter__())



import logging
from typing import Dict, Optional

import numpy as np
from av.codec import CodecContext
from av.error import FFmpegError
from av.packet import Packet
from av.video.codeccontext import VideoCodecContext
from av.video.frame import VideoFrame


class FrameEncoderError(Exception):
    """FrameEncoderError Exception."""

    pass


# TODO: only for testing purpose
# from pathlib import Path
# Path("input").mkdir(parents=True, exist_ok=True)

logger = logging.getLogger("Video frame encoder")


class FrameEncoder:
    """Video frame Encoder."""

    def __init__(
        self, width: int, height: int, fps: float = 30, codec: str = "h264", options: Optional[Dict[str, str]] = None
    ) -> None:
        """Constructor.

        Args:
            width (int): Video frame width.
            height (int): Video frame height.
            fps (float): Video framerate (FPS), default: 30.
            codec (str): Video codec name, e.g. h264, hevc, or vp9, default: h264.
            options (Dict[str, str], optional): Codec options, e.g. {"crf": "0", "preset": "ultrafast",
                "tune": "zerolatency", "x264-params": "keyint=5"}, default: {"preset": "ultrafast",
                "tune": "zerolatency"}.
        """

        # Some default options.
        if options is None:
            if codec == "h264":
                options = {"preset": "ultrafast", "tune": "zerolatency"}
            elif codec == "hevc":
                options = {"crf": "10", "preset": "ultrafast", "tune": "zerolatency", "x265-params": "frame-threads=1"}
            elif codec == "vp9":
                options = {"quality": "realtime", "speed": "8", "lag-in-frames": "0"}
        logger.debug(f"Encoder ({codec}) options: {options}")

        # TODO: only for testing purpose
        # options = {"crf": "0", "preset": "ultrafast", "tune": "zerolatency"}
        # options = {"preset": "ultrafast", "tune": "zerolatency", "x264-params": "keyint=5"}
        # self.frame_id = 0

        self._fps = fps
        self._width = width
        self._height = height
        self._options = options
        self._pix_fmt = "yuv420p"
        self._encoder: VideoCodecContext = CodecContext.create(codec, "w")
        self._init_count = 0
        self.last_timestamp: int = 0
        self._last_frame_is_keyframe = False
        self.encoder_init()

    def width(self) -> int:
        """Get video frame width.

        Returns:
            Video frame width.
        """

        return self._width

    def height(self) -> int:
        """Get video frame height.

        Returns:
            Video frame height.
        """

        return self._height

    def fps(self) -> float:
        """Get video framerate.

        Returns:
            Video framerate.
        """

        return self._fps

    def encoder_init(self) -> None:
        """Init video encoder."""

        self._init_count += 1
        if self._encoder.is_open:
            self._encoder.close()
        self._encoder.width = self._width
        self._encoder.height = self._height
        self._encoder.framerate = self._fps
        self._encoder.pix_fmt = self._pix_fmt
        self._encoder.options = self._options
        self._encoder.open()

    def get_init_count(self) -> int:
        """Get encoder init attempts count.

        Returns:
            Encoder init attempts count.
        """

        return self._init_count

    def last_frame_is_keyframe(self) -> bool:
        """Is last frame a keyframe?

        Returns:
            True if last frame is keyframe.
        """

        return self._last_frame_is_keyframe

    def encode_ndarray(self, frame_data: np.ndarray, format: str = "bgr24") -> bytes:
        """Encode ndarray to packets bytes.

        Args:
            frame_data (ndarray): Video frame / image.
            format (str): Image format.

        Returns:
            Packet data.
        """

        try:
            frame = VideoFrame.from_ndarray(frame_data, format=format)
            # TODO: only for testing purpose
            # frame.to_image().save("input/frame-%04d.jpg" % self.frame_id)
            # self.frame_id += 1

            self._last_frame_is_keyframe = False
            packets = []
            packet: Packet
            for packet in self._encoder.encode(frame):
                # TODO: only for testing purpose
                logger.debug(f"Frame {frame} encoded to packet: {packet}")
                logger.debug(
                    f"packet.pts: {packet.pts}, "
                    f"packet.dts: {packet.dts}, "
                    f"packet.key_frame: {packet.is_keyframe}, "
                    f"packet.is_corrupt: {packet.is_corrupt}"
                )
                if packet.is_keyframe:
                    self._last_frame_is_keyframe = True
                packets.append(bytes(packet))

            # TODO: only for testing purpose
            # if len(packets) == 0:
            #    raise FFmpegError(tag_to_code(b"INDA"), f"No packet encoded from frame {frame}")

            if len(packets) > 1:
                logger.info(f"Frame {frame} encoded to multiple packets: {packets}")
            return b"".join(packets)
        except FFmpegError as e:
            raise FrameEncoderError(e)



import json
import logging
import os
import sys
from threading import Event, Thread
from typing import Callable, Dict, List

import numpy as np
import psutil
import requests

logger = logging.getLogger(__name__)

# Middleware address.
MIDDLEWARE_ADDRESS = str(os.getenv("MIDDLEWARE_ADDRESS", "http://localhost"))
MIDDLEWARE_ADDRESS = MIDDLEWARE_ADDRESS if MIDDLEWARE_ADDRESS.startswith("http") else ("http://" + MIDDLEWARE_ADDRESS)

MIDDLEWARE_REPORT_INTERVAL = float(os.getenv("MIDDLEWARE_REPORT_INTERVAL", 1))

# Event name used for communication between clients and heartbeat module.
HEARTBEAT_CLIENT_EVENT = "heartbeat_client_event"


class LatencyMeasurements:
    """Class for holding data about processing times (latency)."""

    def __init__(self, num_latencies_to_keep: int = 10) -> None:
        self.num_latencies_to_keep = num_latencies_to_keep
        self.processing_latencies = np.zeros(num_latencies_to_keep)

    def store_latency(self, latency: float) -> None:
        """Store latency into "circular" list.

        Args:
            latency (float): Last latency.
        """

        # Remove the oldest entry and add the new one
        # Using strategy "Copy one before and substitute at the end" (fastest)
        self.processing_latencies[0:-1] = self.processing_latencies[1:]
        self.processing_latencies[-1] = latency

    def get_latencies(self) -> List[float]:
        """Get latencies.

        Returns:
            Latencies in list
        """
        latencies: list = self.processing_latencies.tolist()
        return latencies

    def get_avg_latency(self) -> float:
        """Get average latency.

        Returns:
            Average latency.
        """

        return float(np.mean(self.processing_latencies))


class RepeatedTimer(Thread):
    """Repeated interval timer of callback function in Thread."""

    def __init__(self, interval: float, callback: Callable):
        """Constructor.

        Args:
            interval (float): Interval in seconds.
            callback (Callable): Function to be called repeatedly.
        """

        super().__init__(daemon=True)
        self._stop_event = Event()
        self._callback = callback
        self._interval = interval

    def stop(self) -> None:
        """Set stop event to stop FCW worker."""

        self._stop_event.set()

    def run(self):
        """Periodically calls callback function."""

        self._stop_event = Event()
        while not self._stop_event.wait(self._interval):
            self._callback()


class HeartbeatSender:
    """Heartbeat to Middleware sender."""

    def __init__(
        self, status_address: str, heartbeat_data_callback: Callable[[], Dict], repeat_on_error: bool = False
    ) -> None:
        """Constructor, create RepeatedTimer and requests.Session.

        Args:
            status_address (str): Status reporting address.
            heartbeat_data_callback (Callable[[], Dict]): Callback that generates JSON data for heartbeat reporting.
            repeat_on_error (bool): If set, repeat requests further. Defaults to False.
        """

        self.session: requests.Session = requests.Session()
        self.session.headers.update({"Content-type": "application/json"})
        self.connection_error = False
        self.heartbeat_data_callback = heartbeat_data_callback
        self.status_address = status_address
        self.repeat_on_error = repeat_on_error
        self.heartbeat_timer = RepeatedTimer(MIDDLEWARE_REPORT_INTERVAL, self._send_heartbeat_request)
        self.heartbeat_timer.start()

    def _send_heartbeat_request(self) -> None:
        """Send heartbeat request to Middleware."""

        data = self.heartbeat_data_callback()
        if not self.connection_error or self.repeat_on_error:
            logger.debug(f"Sending heartbeat to middleware {self.status_address}: \n{data}")
            try:
                response = self.session.post(self.status_address, json=data, timeout=(1.5, 1.5))
                if response.ok:
                    logger.debug(f"Middleware heartbeat response: \n{self._decode_response(response)}")
                else:
                    logger.warning(
                        f"Middleware heartbeat response: {response}, middleware address: {self.status_address}\n"
                        f"{self._decode_response(response)}"
                    )
                    # self.connection_error = True
            except requests.RequestException as e:
                logger.warning(f"Failed to connect to the middleware address: {self.status_address}, {repr(e)}")
                self.connection_error = True

    def _decode_response(self, response):
        try:
            return json.loads(response.text)
        except json.JSONDecodeError as e:
            logger.warning(f"Cannot decode JSON: {repr(e)}, response: {response.text}")
            return {}


if __name__ == "__main__":
    # Heartbeat sending test.
    logging.basicConfig(
        stream=sys.stdout, level=logging.DEBUG, format="%(asctime)s [%(levelname)s] %(name)s: %(message)s"
    )
    logger = logging.getLogger(__name__)

    def generate_application_heartbeat_data() -> Dict:
        data = {
            "id": "a04dee9a-c303-43b7-8f1b-6c986ad21a99",
            "name": "test",
            "currentRobotsCount": 3,
            "optimalLimit": 5,
            "hardLimit": 10,
        }
        return data

    def generate_robot_heartbeat_data() -> Dict:
        battery_status = psutil.sensors_battery()
        data = {
            "id": "300c719a-1c06-4500-a13a-c2e20592b273",
            "actionSequenceId": None,
            "currentlyExecutedActionIndex": None,
            "batteryLevel": battery_status.percent if battery_status else 100,  # Robot battery level.
            "cpuUtilisation": psutil.cpu_percent(percpu=False),  # Robot CPU utilization.
            "ramUtilisation": psutil.virtual_memory().percent,  # Robot RAM utilization.
        }
        return data

    heartbeat_sender = HeartbeatSender(MIDDLEWARE_ADDRESS + "/status/netapp", generate_application_heartbeat_data)
    heartbeat_sender._send_heartbeat_request()
    del heartbeat_sender
    heartbeat_sender = HeartbeatSender(MIDDLEWARE_ADDRESS + "/status/robot", generate_robot_heartbeat_data)
    heartbeat_sender._send_heartbeat_request()
    del heartbeat_sender



import logging
import time
from typing import Any, Dict

import socketio

from era_5g_interface.channels import DATA_ERROR_EVENT, DATA_NAMESPACE, CallbackInfoClient, Channels, ChannelType

logger = logging.getLogger(__name__)


class ClientChannels(Channels):
    """Channels class is used to define channel data callbacks and contains send functions.

    It handles image frames JPEG, H.264 and HEVC, and JSON LZ4 encoding/decoding. Data is sent via the DATA_NAMESPACE.
    """

    _callbacks_info: Dict[str, CallbackInfoClient]

    def __init__(
        self,
        sio: socketio.Client,
        callbacks_info: Dict[str, CallbackInfoClient],
        **kwargs,
    ):
        """Constructor.

        Args:
            sio (socketio.Client): Socketio Client object.
            callbacks_info (Dict[str, CallbackInfoClient]): Callbacks Info dictionary, key is custom event name.
            disconnect_callback (Callable, optional): Triggered before _shutdown on unhandled exception.
            back_pressure_size (int, optional): Back pressure size - max size of eio.queue.qsize().
            recreate_coder_attempts_count (int): How many times try to recreate the video frame encoder/decoder.
            stats (bool): Store output data sizes.
            extended_measuring (bool): Enable logging of measuring.
        """

        super().__init__(sio, callbacks_info, **kwargs)

        self._sio.on(DATA_ERROR_EVENT, lambda data: self.data_error_callback(data), namespace=DATA_NAMESPACE)

        for event, callback_info in self._callbacks_info.items():
            logger.info(f"Creating client channels callback, type: {callback_info.type}, event: '{event}'")
            if callback_info.type is ChannelType.JSON:
                self._sio.on(
                    event,
                    lambda data, local_event=event: self.json_callback(data, local_event),
                    namespace=DATA_NAMESPACE,
                )
            elif callback_info.type is ChannelType.JSON_LZ4:
                self._sio.on(
                    event,
                    lambda data, local_event=event: self.json_lz4_callback(data, local_event),
                    namespace=DATA_NAMESPACE,
                )
            elif callback_info.type in (ChannelType.JPEG, ChannelType.H264, ChannelType.HEVC):
                self._sio.on(
                    event,
                    lambda data, local_event=event: self.image_callback(data, local_event),
                    namespace=DATA_NAMESPACE,
                )
            else:
                raise ValueError(f"Unknown channel type: {callback_info.type}")

    def json_callback(self, data: Dict[str, Any], event: str) -> Any:
        """Allows to receive general JSON data on DATA_NAMESPACE.

        Args:
            data (Dict[str, Any]): JSON data.
            event (str): Event name.
        """

        timestamp = Channels.get_timestamp_from_data(data)
        cb_info = self._callbacks_info[event]

        try:
            self._input_measuring.log_timestamp(timestamp, "before_callback_timestamp")
            return_value = cb_info.callback(data)
            self._input_measuring.log_timestamp(timestamp, "after_callback_timestamp")
            self._input_measuring.log_measuring(timestamp, "eio_sid", self.get_client_eio_sid(None, DATA_NAMESPACE))
            self._input_measuring.log_measuring(timestamp, "event", event)
            self._input_measuring.store_measuring(timestamp)
            return return_value
        except Exception:
            self._shutdown("JSON", event)

    def json_lz4_callback(self, data: bytes, event: str) -> Any:
        """Allows to receive LZ4 compressed general JSON data on DATA_NAMESPACE.

        Args:
            data (bytes): LZ4 compressed JSON data.
            event (str): Event name.
        """

        before_lz4_decode_timestamp = time.perf_counter_ns()
        decoded_data = super().data_lz4_decode(data, event)
        if decoded_data:
            after_lz4_decode_timestamp = time.perf_counter_ns()
            timestamp = Channels.get_timestamp_from_data(decoded_data, default=before_lz4_decode_timestamp)
            self._input_measuring.log_measuring(timestamp, "before_lz4_decode_timestamp", before_lz4_decode_timestamp)
            self._input_measuring.log_measuring(timestamp, "after_lz4_decode_timestamp", after_lz4_decode_timestamp)
            return self.json_callback(decoded_data, event)

    def image_callback(self, data: Dict[str, Any], event: str) -> Any:
        """Allows to receive JPEG or H.264 or HEVC encoded image on DATA_NAMESPACE.

        Args:
            data (Dict[str, Any]): Received dictionary with frame data.
            event (str): Event name.
        """

        timestamp = Channels.get_timestamp_from_data(data)
        cb_info = self._callbacks_info[event]

        self._input_measuring.log_timestamp(timestamp, "before_decode_timestamp")
        decoded_data = super().image_decode(data, event)
        self._input_measuring.log_timestamp(timestamp, "after_decode_timestamp")

        if decoded_data:
            try:
                self._input_measuring.log_timestamp(timestamp, "before_callback_timestamp")
                return_value = cb_info.callback(decoded_data)
                self._input_measuring.log_timestamp(timestamp, "after_callback_timestamp")
                self._input_measuring.log_measuring(timestamp, "eio_sid", self.get_client_eio_sid(None, DATA_NAMESPACE))
                self._input_measuring.log_measuring(timestamp, "event", event)
                self._input_measuring.store_measuring(timestamp)
                return return_value
            except Exception:
                self._shutdown("image", event)
